(* $Id: fw.ml,v 1.15 2005/06/27 10:46:24 regisgia Exp $ *)

open Topology

open Misc.F_INT
open Misc

type 'a ptype_info = {
  mutable pti_car  : 'a;
  mutable pti_foot : 'a;
}

(* -------------- *)
(* Floyd Warshall *)
(* -------------- *)
type fw_base = {
  fw_weight : f_int array array;
  fw_ngh    : int array array;
}

type fw = {
  fw_pathes : fw_base ptype_info;
  fw_hq     : int;
}

type fw_next_step = {
  fns_next        : int;
  fns_ptype_one   : Message.ptype;
  fns_ptype_final : Message.ptype;
}

let create_empty_fw = fun n_nodes ->
  { fw_weight = Array.create_matrix n_nodes n_nodes F_Inf;
    fw_ngh    = Array.create_matrix n_nodes n_nodes (-1); }

let create_fw = fun n_nodes ngh ->
  let fw = create_empty_fw n_nodes in
    for node_idx = 0 to (n_nodes-1) do
      (* Création de la matrice d'adjacences *)
      List.iter
        (fun ngh_idx ->
           assert (node_idx <> ngh_idx);
           fw.fw_weight.(node_idx).(ngh_idx) <- (F_Int 1);
           fw.fw_ngh.(node_idx).(ngh_idx) <- node_idx)
        ngh.(node_idx);

      (* Boucles sur le même noeud *)
      fw.fw_weight.(node_idx).(node_idx) <- F_Int 0
    done;
    fw

let init_fw = fun topo ->
  (* Floyd-Warshall *)
  let rec fw_iter = fun fw fw_out k_iter i j ->
    let d1 = fw.fw_weight.(i).(j)
    and d2 = fw.fw_weight.(i).(k_iter) ++ fw.fw_weight.(k_iter).(j)
    in
      fw_out.fw_ngh.(i).(j) <-
        if   (<<=) d1 d2
        then fw.fw_ngh.(i).(j)
        else fw.fw_ngh.(k_iter).(j);
      fw_out.fw_weight.(i).(j) <- f_min d1 d2
  in

  let n_nodes = nodes_count topo in

  let fw_car      = ref (create_fw n_nodes (topo.t_neigh_c))
  and fw_foot     = ref (create_fw n_nodes (topo.t_neigh_f))
  and fw_car_tmp  = ref (create_empty_fw n_nodes)
  and fw_foot_tmp = ref (create_empty_fw n_nodes)
  in
    for k = 0 to n_nodes-1 do
      for i = 0 to n_nodes-1 do
        for j = 0 to n_nodes-1 do
          fw_iter !fw_car  !fw_car_tmp  k i j;
          fw_iter !fw_foot !fw_foot_tmp k i j
        done
      done;
      Misc.swap fw_car  fw_car_tmp;
      Misc.swap fw_foot fw_foot_tmp
    done;
    { fw_pathes = { pti_car = !fw_car; pti_foot = !fw_foot };
      fw_hq     = (get_node topo "55-and-woodlawn").Message.nod_num; }

let extract_pathes = fun fw -> function
  | Message.Cop_car  -> fw.fw_pathes.pti_car
  | Message.Cop_foot -> fw.fw_pathes.pti_foot
  | _ -> invalid_arg "extract_data (Cop_car, Cop_foot)"

let get_shortest_path_infos = fun fw ptype ~node_from ~node_to ->
  let this_pathes = extract_pathes fw ptype
  and neg_pathes  = extract_pathes fw (Message.neg_ptype ptype)
  in

  let direct_dist = this_pathes.fw_weight.(node_from).(node_to)
  and neg_dist    =
    this_pathes.fw_weight.(node_from).(fw.fw_hq)
    ++ neg_pathes.fw_weight.(fw.fw_hq).(node_to)

  in
    if (<<=) direct_dist neg_dist then
      (direct_dist, ptype, ptype)
    else begin
      if node_from = fw.fw_hq then
        (neg_dist, (Message.neg_ptype ptype), (Message.neg_ptype ptype))
      else
        (neg_dist, ptype, (Message.neg_ptype ptype))
    end

let shortest_path_next_step = fun fw ptype ~node_from ~node_to ->
  let rec find_next_step = fun pathes ?(current = None) node_from node_to ->
    if node_from = node_to then
      current
    else let prev = pathes.(node_from).(node_to) in
      if   prev < 0
      then None
      else find_next_step pathes ~current:(Some node_to) node_from prev
  in

  let create_fw_result = fun next ptype_1 ptype_final ->
    match next with
      | None -> None
      | Some next ->
          Some { fns_next        = next;
                 fns_ptype_one   = ptype_1;
                 fns_ptype_final = ptype_final; }
  in

  let (dist, ptype_1, ptype_final) =
    get_shortest_path_infos fw ptype ~node_from ~node_to
  in

  let use_hq   = ptype_final <> ptype
  and this_pathes = extract_pathes fw ptype
  and neg_pathes  = extract_pathes fw (Message.neg_ptype ptype)
  in

    if use_hq then begin
      if node_from = fw.fw_hq then
        (dist,
         create_fw_result
           (find_next_step neg_pathes.fw_ngh fw.fw_hq node_to)
           (Message.neg_ptype ptype) (Message.neg_ptype ptype))
      else
        (dist,
         create_fw_result
           (find_next_step this_pathes.fw_ngh node_from fw.fw_hq)
           ptype (Message.neg_ptype ptype))
    end else
      (dist,
       create_fw_result
         (find_next_step this_pathes.fw_ngh node_from node_to)
         ptype ptype)

let dump_shortest_path = fun fw topo ptype ~node_from ~node_to ->
  let rec dump = fun pathes ?(acc = []) node_from node_to ->
    if node_from = node_to then
      Some (List.rev acc)
    else let prev = pathes.(node_from).(node_to) in
      if   prev < 0
      then None
      else dump pathes ~acc:(node_to :: acc) node_from prev
  in

  let (dist, ptype_1, ptype_final) =
    get_shortest_path_infos fw ptype ~node_from ~node_to
  in

  let use_hq   = ptype_final <> ptype
  and this_pathes = extract_pathes fw ptype
  and neg_pathes  = extract_pathes fw (Message.neg_ptype ptype)
  in

  let nodes =
    if use_hq then
      let d1 = dump this_pathes.fw_ngh node_from fw.fw_hq
      and d2 = dump neg_pathes.fw_ngh fw.fw_hq node_to
      in match d1, d2 with
        | Some d1, Some d2 ->
            let d1 = List.map (fun x -> (x, ptype)) d1
            and d2 = List.map (fun x -> (x, Message.neg_ptype ptype)) d2
            in
              Some (List.append d1 d2)
        | _ -> None
    else
      Misc.apply_opt
        (List.map (fun x -> (x, ptype)))
        (dump this_pathes.fw_ngh node_from node_to)
  in
    match nodes with
      | None ->
          Printf.sprintf "%d -> %d : NONE" node_from node_to
      | Some nodes ->
          Printf.sprintf "%d -> %d : %d (%s)"
            node_from node_to
            (int_of_fint dist)
            (String.concat "; "
               (List.map
                  (fun (x, y) ->
                     Printf.sprintf "%s/%s"
                       (topo.Topology.t_nods.(x).Message.nod_loc)
                       (Message.ptype_to_string y))
                  nodes))

(* Calcul des positions potentiellement occupables par un voleur au
 * bout d'un nombre donné de tours *)
type witr_cont = {
  witr_topo   : topo;
  witr_robber : IntSet.t;
}

let witr_cont = fun topo ~robber_loc ->
  if robber_loc < 0 || robber_loc >= nodes_count topo then
    invalid_arg "wirt_conf: robber_loc";
  { witr_topo = topo; witr_robber = IntSet.singleton robber_loc }

let where_is_the_robber = fun ~cops ~nturns witr_cont ->
  if nturns < 0 then
    invalid_arg "where_is_the_robber: nturns < 0";
  let rec iter = fun n robbers ->
    let robbers = IntSet.diff robbers cops in
      if n >= nturns then
        robbers
      else begin
        let new_locs =
          IntSet.fold
            (fun robber_loc locs ->
               List.fold_left
                 (fun locs node -> IntSet.add node locs)
                 (IntSet.add robber_loc locs)
                 witr_cont.witr_topo.t_neigh_f.(robber_loc))
            IntSet.empty robbers
        in
          iter (n+1) new_locs
      end
  in
    { witr_cont with witr_robber = iter 0 witr_cont.witr_robber }

let where_is_the_robber_cops_are_moving = fun ~cops witr_cont ->
  List.fold_left
    (fun witr_cont these_cops ->
       where_is_the_robber ~cops:these_cops ~nturns:1 witr_cont)
    witr_cont cops

let witr_topo   = fun { witr_topo = topo } -> topo
let witr_robber = fun { witr_robber = robber } -> robber
