(* $Id: info.ml,v 1.29.2.10 2005/07/10 13:54:00 frisch Exp $ *)

open Message;;
open Printf;;
open Topology;;

(* Il y a une structure info par monde observ.
   Elle est cre  partir de la structure prcdente
   en ajoutant les infos dduites du monde courant.

   On la met  jour en place pour intgrer les informations
   fournies par les autres flics. *)

type info = {
  now: int; (* toujours impair ! *)
  brains: Brains.map; 
  (* correspondence flic->joueur.
     C'est une variable globale !
     Quand un joueur A prend le controle sur un joueur B,
     on oublie ses informations: si elles taient fausses, on risquerait
     de dire que A est menteur. *)
     
  history: info option; (*  now = world + 2 *)
  unsure : NS.t array; 
  (* Pour chaque **joueur**, l'ensemble de prsence
     possible d'aprs les infos qu'il a donnes.
     Pour les flics control, la case est inutilise. *)

  trusted : bool array; 
  (* Pour chaque **joueur**: est-ce qu'on le croit ?
     Note: trusted est une variale globale !
     Pour les flics controls, trusted = false. *)

  mutable summary: NS.t option; (* Intersection de:
			   - extend(summary precedent)
			   - unsure.(i) si trusted.(i)
			   Mettre  None lorsque trusted change
			   ou lorsque les infos passes (ou unsure) ont chang,
			   pour forcer recalcul *)

  topo : topo;
  display : Draw.t;
  cops : (nod * ptype) array;

  robber_loc : nod option;

  players : int list array;
  (* Pour chaque joueur, la liste des flics controls.
     Pour les autres case, liste vide. *)

  me : int;

  dirty_brains: int list;
  (* Liste des joueurs corrompus.
     Non vide <==> Je suis corrompu. *)

  ripoux: int list ref;
  (* Variable globale ! *)

  robbed: int; 
}

let topo info = info.topo
let drawer info = info.display


let am_i_dirty info = info.dirty_brains <> []
let brains info = info.brains

let get_my_cops info = info.players.(info.me)
let get_other_cops info =
  List.fold_left 
    (fun cops b ->
       if b <> info.me then info.players.(b) @ cops
       else cops) []
    (Brains.brains info.brains)

let current_world info =
  info.now

let plan_for_world info =
  assert(info.now > 0);
  succ info.now

let init topo display =
  let ini = NS.singleton topo.initial_robber in
  let brains = Brains.init topo.cops_name in
  { topo = topo;
    display = display;
    now = -1; 
    brains = brains;
    history = None; 
    summary = Some ini;
    unsure = Array.create topo.ncops ini;
    trusted = Array.create topo.ncops true;

    cops = [||]; (* cops,players: seront remplis par add_world *)
    players = [||];

    me = Brains.cop_num brains topo.name;
    dirty_brains = [];
    ripoux = ref [];
    robber_loc = None;
    robbed = 0;
  }

let find_bot w name =
  try List.find (fun pl -> pl.pl_bot = name) w.wor_pls
  with Not_found -> raise Not_found

let get_cop_i info i = info.cops.(i)
(*
let get_cop info s = info.cops.(cop_number info.topo s)
let get_me info = get_cop info info.topo.name
*)

let rec invalidate_summary = function
  | None -> ()
  | Some i -> i.summary <- None; invalidate_summary i.history

let rec summary info =
  match info.summary with
    | Some s -> s
    | None ->
	(* Vrifier si l'un des flics est un menteur *)
	let me_info = info.unsure.(info.me) in
	let untrust = ref false in
	Array.iteri
	  (fun i trust -> 
	     info.unsure.(i) <- NS.inter info.unsure.(i) me_info;
	     if trust && NS.is_empty info.unsure.(i) then
	       (info.trusted.(i) <- false; untrust := true;
		Printf.eprintf "Cop %i is a LIAR !\n" i))
	  info.trusted;
	if !untrust then invalidate_summary info.history;

	let prev =
	  match info.history with
	    | None -> NS.singleton info.topo.initial_robber
	    | Some i -> summary i in
	let s = ref (extend info.topo prev) in
	Array.iteri 
	  (fun i trust -> if trust then s := NS.inter !s info.unsure.(i))  
	  info.trusted;

	(* Meme si aucun flic n'est detect comme menteur, 
	   il se peut que globalement, l'intersection soit vide.
	   Dans ce cas, on utilise seulement l'information de notre flic.
	   Normalement, a ne peut pas tre vide (sinon c'est un bug). *)

	if NS.is_empty !s then s := info.unsure.(info.me);

	info.summary <- Some !s;
	!s

let ptype_to_string p =
  match p with
  | Cop_car -> "cop-car"
  | Cop_foot -> "cop-foot"
  | Robber -> "robber"
;;

let print_one_inf i =
  eprintf "inf: %s %s %s %d %d\n" i.inf_bot i.inf_loc
         (ptype_to_string i.inf_ptype) i.inf_world i.inf_certainty;
;;

let dump_info info s =
(*
  let sum = summary info in
  Format.eprintf "Time %i (%s): " info.now s;
  Array.iter (fun s -> Format.eprintf "%i " (NS.cardinal s)) info.unsure;
  Format.eprintf " --> %i\n" (NS.cardinal sum);
  flush stderr
*)
  ()


(* At time ev_world, the robber is spotted at ev_loc. *)
let integrate_evidence info0 detect_scoundrel cop_name ev_world ev_loc =
  let topo = info0.topo in
  let player = Brains.controller info0.brains cop_name in
  Printf.eprintf 
    "Integrating evidence at time %i, loc:%s, for cop %s\n" 
    ev_world ev_loc.nod_loc cop_name;

  (* Go back to the past to integrate this new info. *)
  let rec go_back info = 
    if info.now = ev_world then 
      let loc = NS.singleton ev_loc in
      info.unsure.(player) <- NS.inter loc info.unsure.(player);
      info.summary <- None
    else match info.history with
      | Some old ->
	  (* Est-ce qu'il y avait un policier  cet endroit ?
	     Si oui, c'tait un ripoux ! *)
	  if detect_scoundrel then (
	    Array.iteri
	      (fun cop_i (nod,_) ->
		 if nod = ev_loc then 
		   Array.iteri
		     (fun brain_i cops ->
			if List.mem cop_i cops && 
			  not (List.mem brain_i !(info.ripoux)) then (
			    eprintf "Ripoux detected: %i\n" brain_i;
			    info.ripoux := brain_i :: !(info.ripoux)
			  )
		     ) info.players
	      ) info.cops;
	  );

	  go_back old;
	  info.summary <- None;
	  info.unsure.(player) <- 
	    NS.inter info.unsure.(player) (extend topo old.unsure.(player))
      | _ -> 
	  Printf.eprintf 
	    "CANNOT FIND OLD INFO TO INTEGRATE EVIDENCE!!!\n"
  in
  go_back info0


(* Prendre en compte les infos donnes par un autre flic *)
let add_inf info (name, infs) =
(*
  eprintf "report from %s:\n" name;
  let cmp i1 i2 = compare i1.inf_loc i2.inf_loc in
  List.iter print_one_inf (List.sort cmp infs);
  flush stderr;
*)
  let topo = info.topo in
  let player = Brains.controller info.brains name in
  (* Normalement, c'est dj le nom du joueur que l'on recupere ... *)

  let f i =
    let nod = get_node topo i.inf_loc in
    if i.inf_bot = topo.robber_name
       && i.inf_ptype = Robber
       && (i.inf_world = info.now || i.inf_world = pred info.now)
       && i.inf_certainty = -100
    then begin
      (* On nous dit que le voleur n'est pas l, maintenant. *)
(*      Printf.eprintf "Robber is not at %s, now\n" i.inf_loc; *)
      info.unsure.(player) <- NS.remove nod info.unsure.(player);
    end else if i.inf_bot = topo.robber_name
      && i.inf_ptype = Robber
      && i.inf_certainty = 100
      && i.inf_world <= info.now
    then begin
      (* On nous dit que le voleur a t reper dans le pass *)
      let t = if i.inf_world mod 2 = 0 then i.inf_world - 1
      else i.inf_world in
      integrate_evidence info false name t nod
    end else begin
      Printf.eprintf 
	"Unused information from %s about %s, certainty %i, at %s, world %i (currently world=%i)\n"
	name
	i.inf_bot
	i.inf_certainty
	i.inf_loc
	i.inf_world
	info.now
    end
  in
  if info.trusted.(player) then List.iter f infs

(*
and integrate_absence_of_evidence_sure info0 ev_loc =
  let topo = info0.topo in
  Printf.eprintf 
    "Integrating absence of evidence at time %i, loc:%s\n" 
	info0.now ev_loc.nod_loc;
  (* Go back to the past to integrate this new info.
     Return a new replacement info or raise Exit if no change. *)
  let rec go_back info = 
    if not (info.now = 0 || info0.now - info.now >= 24) then info
      else
    (* Too late *)

    let info = 
      match info.world,info.informs with
	| Some (w,old), Some informs ->
	    let old' = go_back old in
	    if old == old' then info
	    else (
	      Printf.eprintf "Replay time %i\n" info.now;
	      let fut = add_world_noevidence old w in
	      let fut = add_informs_noevidence fut informs in
	      { fut with summary = NS.inter fut.summary info.summary })
	| _ -> info in

    (* Is there a cop here ? *)
    if 
      (try
	 Array.iter 
	   (fun (n,_) -> if n.nod_num = ev_loc.nod_num then raise Exit)
	   info.cops;
	 false
       with Exit -> true) then info
    else
      (* No, let's continue *)

      if (info.now mod 8 = 1 && info.now > 1) then
	(* If the robber was here at that time, there would be a mark. *)
	if not (NS.mem ev_loc info.summary) || not (NS.mem ev_loc info.sure) 
	then
	  { info with 
	      sure = NS.remove ev_loc info.sure;
	      summary = NS.remove ev_loc info.summary }
	else info
      else info
  in
  let info0' = go_back info0 in
  flush stderr;
  if not (NS.equal info0'.summary info0.summary) then
    Draw.display_node_sets info0.display ~title:"Reducing by negative evidence" ~wait:true
      info0.summary info0'.summary;
  info0'
*)


let integrate_my_evidences info0 w =
  let topo = info0.topo in
  (* Collect of my evidences *)
  let evs = 
    List.map (fun e -> e.ev_world, get_node topo e.ev_loc) w.wor_evs in
  (* Only the most recent one is useful: exact information *)
  match List.sort (fun (t1,_) (t2,_) -> t2 - t1) evs with
    | (ev_world,ev_loc)::_ when pred ev_world <  w.wor_wor ->
	(* The info is also about the world -1 *)
	let ev_world = pred ev_world in
	(* J'utilise mon nom propre: ok, vu que c'est moi qui controle
	   tous mes flics. *)
	integrate_evidence info0 true topo.name ev_world ev_loc
(*    | [] ->
	integrate_absence_of_evidence_sure info0	  (get_node info0.topo (find_bot w info0.topo.name).pl_loc) *)
    | _ -> ()


(* Intgrer les infos du monde que l'on vient d'observer,
   crer une nouvelle structure *)
let add_world info0 w =
  (* Prendre en compte les prises de controle. *)
  let taking = Brains.integrate info0.brains w.wor_sc in
  List.iter 
    (fun (perd,gagne) -> 
       Printf.eprintf "Taking over detected %i->%i\n" perd gagne;
       info0.trusted.(perd) <- false)
    taking;

  if taking <> [] then invalidate_summary (Some info0);
  
  (* Intgrer d'abord dans le pass les traces que j'ai trouves. *)
  integrate_my_evidences info0 w;
  dump_info info0 "after my evidences";

  let topo = info0.topo in

  (* Extraire les autres infos du monde courant. *)
  let now = w.wor_wor in

  let robber_loc = 
    try Some (get_node topo (find_bot w topo.robber_name).pl_loc)
    with Not_found -> None in

  let cops = 
    Array.map
      (fun cn ->
	 let pl = find_bot w cn in
	 (get_node topo pl.pl_loc, pl.pl_ptype)
      ) topo.cops_name in

  let players =
    Array.init 
      (Array.length topo.cops_name)
      (fun i -> Brains.controled_by info0.brains i) in

  let dirty = 
    List.map (Brains.cop_num info0.brains) w.wor_dc in

  let my_cops_num = players.(info0.me) in
  let my_cops = List.map (fun i -> cops.(i)) my_cops_num in
  let my_info = ref (extend topo info0.unsure.(info0.me)) in
  if w.wor_smell = 0 then
    my_info := NS.diff !my_info (player_nosmell topo my_cops)
  else
    my_info := NS.inter !my_info (player_smell topo my_cops w.wor_smell);

  if dirty = [] then (
    (* ne pas faire ca si je suis corrompu. *)
    let robber_found = ref false in
    let f p =
      let nod = get_node topo p.pl_loc in
      if p.pl_bot = topo.robber_name then begin
	my_info := NS.inter !my_info (NS.singleton nod);
	robber_found := true;
      end else begin
	my_info := NS.remove nod !my_info;
      end
    in
    List.iter f w.wor_pls;
  

  (* Le voleur n'a pas t trouv sur une banque. Donc on sait
     qu'il n'est sur aucune banque. *)
    if not !robber_found then my_info := NS.diff !my_info topo.t_banks;
  );

  let unsure = Array.mapi (fun i x -> 
			     if info0.trusted.(i) 
			     then NS.inter (extend topo x) !my_info
			     else x) info0.unsure 
  in
  let info = 
    { now = now;
      history = Some info0;
      summary = None;
      topo = topo;
      cops = cops;
      players = players;
      trusted = info0.trusted;
      brains = info0.brains;
      unsure = unsure;
      display = info0.display;
      me = info0.me;
      dirty_brains = dirty;
      ripoux = info0.ripoux;
      robber_loc = robber_loc;
      robbed = w.wor_rbd;
    } in
  dump_info info "after add_world";
  info

(* Mettre  jour la structure courante pour prendre en compte
   les infos des autres flics. *)
let add_informs info l =
  List.iter (add_inf info) l;
  dump_info info "after add_informs"

let debug_trusted info = Array.iter (eprintf "trusted = %B\n") info.trusted;;


let get_possible info = summary info
let get_certain info = info.unsure.(info.me)


let map_cops info f =
  List.map 
    (fun (cop_name,i) -> 
       let w,t = info.cops.(i) in
       f cop_name i w t)
    info.topo.cops_number

let iter_cops info f =
  ignore (map_cops info f : unit list)

let change_cop info s =
  let cops' = Array.copy info.cops in
    List.iter (fun (i, ptype) ->
                let loc, old_type = cops'.(i) in
                  assert (old_type <> ptype);
                  assert (loc.nod_tag = Hq);
                  cops'.(i) <- (loc, ptype)) s;
    { info with cops = cops' }


let get_ripoux info =
  List.filter (Brains.is_active info.brains) !(info.ripoux)

let get_robber info =
  info.robber_loc

let robbed info =
  info.robbed
