(* $Id: planner.ml,v 1.22.4.1 2005/07/09 21:36:40 frisch Exp $ *)

open Printf
open Message
open Topology
open Info

(* Magic coefficients *)

let potential_Bank = 50                 (* static bonus for a bank *)
let potential_neighbour = 0             (* static bonus for every neighbour *)
let potential_not_here_certain = 0      (* we're certain robber isn't here *)
let potential_not_here_maybe = 0 (*5*)       (* we believe robber isn't here *)
let potential_maybe_here = 10000         (* robber could be here *)
let potential_last_visited = 200      (* max bonus for rarely visited loc *)
let repulsion_cops = -100.0

let score_acceptable_thr = 0.1

type state = {
  topo: topo;
  drawer: Draw.t;
  pot_car: int array;
  pot_foot: int array;
  last_visited: int array
}

(* Compute static potentials for each node.
   What contributes to the static potential of a node:
     - the tag of the node (bonus for Bank)
     - the connectivity (number of neighbours).
*)

let potential_node_tag = function
  | Bank -> potential_Bank
  | _    -> 0

let init topo drawer =
  let potential_car n =
    potential_node_tag n.nod_tag +
    potential_neighbour * List.length topo.t_neigh_c.(n.nod_num) in
  let potential_foot n =
    potential_node_tag n.nod_tag +
    potential_neighbour * List.length topo.t_neigh_f.(n.nod_num) in
  { topo = topo;
    drawer = drawer;
    pot_car = Array.map potential_car topo.t_nods;
    pot_foot = Array.map potential_foot topo.t_nods;
    last_visited = Array.make (Array.length topo.t_nods) (-1) }

(* The dynamic potential of a node is its static potential plus
      - potential_not_here_certain if we are certain the robber isn't there
      - potential_not_here_maybe if information from other cops
        tell us it isn't there
      - otherwise, it's potential_maybe_here scaled according to
        connected component analysis
   times a bonus for points that are at the frontier of the possible 
             robber positions
   plus the number of rounds it hasn't been visited (capped) *)

type coefs = { static : int; robber : int; lastv : int; repulse : int };;
let cc = { static = 1; robber = 1; lastv = 1; repulse = 1 };;

let potential_node sd connect certain possible now ty n c =
  let static_pot =
    match ty with
    | Cop_car -> sd.pot_car.(n.nod_num)
    | Cop_foot -> sd.pot_foot.(n.nod_num)
    | Robber -> assert false in
  let robber_pot =
    if not (NS.mem n certain) then potential_not_here_certain else
    if not (NS.mem n possible) then potential_not_here_maybe else
    Connected.split connect potential_maybe_here n in
  let robber_pot2 =
    let neigh = possible_dest sd.topo n Robber in
    robber_pot * (1 + NS.cardinal (NS.diff neigh possible))
  in
  let last_visited_pot =
    min potential_last_visited (now - sd.last_visited.(n.nod_num)) in
  c.static * static_pot + c.robber * robber_pot2 + c.lastv * last_visited_pot

(* The potential field at a point A is

                             potential(B)
     field(A) = sum_{B node} ------------
                             1 + d(A, B)^2

   We are interested in the gradient of this field, that is

     grad(A,A') = field(A') - field(A)

                               potential(B)    potential(B)
                = sum_{B node} ------------  - ------------
                               1 + d(A', B)^2  1 + d(A, B)^2

                               potential(B) * (d(A, B)^2 - d(A', B)^2)
                = sum_{B node} -----------------------------------
                                  (1 + d(A', B)^2) * (1 + d(A, B)^2)
*)

let dist sd ty a b =
  match ty with
  | Cop_car  -> dist_c sd.topo a b
  | Cop_foot -> dist_f sd.topo a b
  | Robber   -> assert false

let gradient sd connect certain possible now ty a a' c =
  let add_node b g =
    let dab = (let d = dist sd ty a b in d * d)
    and da'b = (let d = dist sd ty a' b in d * d) in
    g +.
      float (potential_node sd connect certain possible now ty b c
             * (dab - da'b))
      /.
      float ((1 + da'b) * (1 + dab)) in
  Array.fold_right add_node sd.topo.t_nods 0.0

(* We add another field that corresponds to repulsion between cops.

                                        repulsion_cops
     repfield(A) = sum_{B earlier cops} --------------
                                        1 + d(A, B)^2

   Again, we are interested in the gradient of this field, that is

     repgrad(A,A') = repfield(A') - repfield(A)

                                           d(A, B)^2 - d(A', B)^2
   = repulsion_cops * sum_{B earlier cops} ------------------------------
                                           (1 + d(A', B)^2) * (1 + d(A, B)^2)

   Note that repulsion_cops < 0 for this to be effective. *)

let repulsive_gradient sd othercops ty a a' =
  let add_cop b g =
    let dab2 = (let d = dist sd ty a b in d * d) 
    and da'b2 = (let d = dist sd ty a' b in d * d) in
    g +. float (dab2 - da'b2) /. (float (1 + da'b2) *. float (1 + dab2)) in
  repulsion_cops *. List.fold_right add_cop othercops 0.0

(* Compute and score all possible moves A' for a cop at A.
   A' ranges over the nodes reachable in 0 or 1 moves from A.
   The score is grad(A, A') + repgrad(A, A'). *)

let scored_moves sd connect certain possible now ty othercops a =
  let n_zones = Connected.classes connect in
  let card = NS.cardinal possible in
  let coefs =
    if card < 5 then
      { static = 0; robber = 1; lastv = 0; repulse = 1 }
    else if card > 50 then
      { static = 0; robber = 0; lastv = 10; repulse = 10 }
    else if n_zones = 1 then
      { static = 1; robber = 1; lastv = 0; repulse = 5 }
    else
      { static = 1; robber = 1; lastv = 1; repulse = 8 }
  in
(*
  eprintf "mode: static=%d robber=%d lastv=%d repulse=%d\n"
          coefs.static coefs.robber coefs.lastv coefs.repulse;
*)
  let neigh =
    match ty with
    | Cop_car  -> sd.topo.t_neigh_c.(a.nod_num)
    | Cop_foot -> sd.topo.t_neigh_f.(a.nod_num)
    | Robber   -> assert false in
(*
  eprintf "Move from %s to:\n" a.nod_loc;
*)
  let score_move n =
    let a' = sd.topo.t_nods.(n) in
    let g1 = gradient sd connect certain possible now ty a a' coefs
    and g2 = float coefs.repulse *. repulsive_gradient sd othercops ty a a' in
(*
    eprintf "  %.3e  %.3e %.3e  %s\n" g1 g2 (g1 +. g2) a'.nod_loc;
*)
    (a', g1 +. g2) in
  (a, 0.0) :: List.map score_move neigh

(* Select highest-scored move in the given list *)

let preferred_move scores =
  match scores with
  | [] -> assert false
  | (a0, sc0) :: rem ->
      let best_move = ref a0 and best_score = ref sc0 in
      List.iter
        (fun (a, sc) ->
          if sc >= !best_score then (best_score := sc; best_move := a))
        scores;
      !best_move

(* Construct the planning.
   Return: - the scoring of our possible moves
           - the plan for all bots *)

let find_bot w name =
  try List.find (fun pl -> pl.pl_bot = name) w.wor_pls
  with Not_found -> assert false

type scores = (int * ((nod * float) list)) list

let make_plan sd info certain possible =
  let now = current_world info in
  (* Update last_visited *)
  Info.iter_cops info
    (fun name n loc ptype ->
       let f nod = sd.last_visited.(nod.nod_num) <- now in
       NS.iter f (nosmell sd.topo loc ptype);
       f loc);

  let connect = Connected.decompose sd.topo possible in
  (* DEBUG *)
  Array.iter
    (fun n ->
       if Connected.repr connect n >= 0 then
       Draw.draw_label sd.drawer n
	 (string_of_int (Connected.repr connect n))
(*	 (string_of_int (potential_node sd connect certain possible
                             w.wor_wor me.pl_ptype n cc)) *)
    )
    sd.topo.t_nods;
  

  let my_cops = Info.get_my_cops info in
  let other_cops = Info.get_other_cops info in
  let all_scores = ref [] in
  let plan = ref [] in
  let moves = ref [] in

  let best_move i =
    let loc,ptype = Info.get_cop_i info i in
    let scores = scored_moves sd connect certain possible now ptype !moves loc in
    let move = preferred_move scores in
    all_scores := (i,scores) :: !all_scores;
    plan := (Planify.mk_plan info sd.topo.cops_name.(i) ptype move) :: !plan;
    moves := move :: !moves in

  List.iter best_move (Info.get_my_cops info);
  List.iter best_move (Info.get_other_cops info);
  !all_scores,!plan

(* Decide which move we will do.
   Select the first move proposed for us in the winning vote
   provided its score is >= score_acceptable_thr * highest score.
   If none is found, select move with highest score. *)

exception InvalidPlan

let decide_move_for sd info scores plan i = 
  let loc,ptype = Info.get_cop_i info i in
  let scores = List.assoc i scores in
  let best_score = List.fold_left (fun bsc (a, sc) -> max bsc sc) 0.0 scores in
  let me_name = sd.topo.cops_name.(i) in
  try
  let suggested =
    try List.find 
      (fun pl -> pl.plan_bot = me_name && pl.plan_world = 1 + current_world info)
      plan 
    with Not_found -> raise InvalidPlan in
  let suggested =
    try get_node sd.topo suggested.plan_loc
    with _ -> raise InvalidPlan
  in
  let score =
    try List.assoc suggested scores 
    with Not_found -> raise InvalidPlan in
  if score >= score_acceptable_thr *. best_score
  then (eprintf "Selecting voted dest %s\n" suggested.nod_loc; (ptype, suggested))
  else (ptype, preferred_move scores)
  with InvalidPlan -> (ptype, preferred_move scores)

let decide_move sd info scores plan =
  List.map (fun i -> i,decide_move_for sd info scores plan i) (Info.get_my_cops info)

